
c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine x_solve

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     
c     Performs line solves in X direction by first factoring
c     the block-tridiagonal matrix into an upper triangular matrix, 
c     and then performing back substitution to solve for the unknow
c     vectors of each line.  
c     
c     Make sure we treat elements zero to cell_size in the direction
c     of the sweep.
c     
c---------------------------------------------------------------------

      include 'header.h'

      call lhsx
      call x_solve_cell
      call x_backsubstitute

      return
      end
      
      
      subroutine x_backsubstitute

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     back solve: if last cell, then generate U(isize)=rhs(isize)
c     else assume U(isize) is loaded in un pack backsub_info
c     so just use it
c     after call u(istart) will be sent to next cell
c---------------------------------------------------------------------

      include 'header.h'

      integer i, j, k, m, n

      do k=1,grid_points(3)-2
         do j=1,grid_points(2)-2
            do i=grid_points(1)-2,0,-1
               do m=1,BLOCK_SIZE
                  do n=1,BLOCK_SIZE
                     rhs(m,i,j,k) = rhs(m,i,j,k) 
     >                    - lhs(m,n,cc,i,j,k)*rhs(n,i+1,j,k)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return
      end


c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine x_solve_cell

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     performs guaussian elimination on this cell.
c     
c     assumes that unpacking routines for non-first cells 
c     preload C' and rhs' from previous cell.
c     
c     assumed send happens outside this routine, but that
c     c'(IMAX) and rhs'(IMAX) will be sent to next cell
c---------------------------------------------------------------------

      include 'header.h'

      integer i,j,k,isize

      double precision ablock(5,5,0:JMAX/2*2), bblock(5,5,0:JMAX/2*2),
     >                 cblock(5,5,0:JMAX/2*2), dblock(5,5,0:JMAX/2*2)
      double precision avec(5,0:JMAX/2*2), bvec(5,0:JMAX/2*2)
!HPF$ sequence ablock, bblock, cblock, dblock, avec, bvec

      integer ii, jj

      interface
         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine matvec_sub(ablock,avec,bvec)
         double precision ablock,avec,bvec
         dimension ablock(5,5),avec(5),bvec(5)
!HPF$    sequence ablock, avec, bvec
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine matmul_sub(ablock, bblock, cblock)
         double precision ablock, bblock, cblock
         dimension ablock(5,5), bblock(5,5), cblock(5,5)
!HPF$    sequence ablock, bblock, cblock
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine binvcrhs( lhs,c,r )
         double precision lhs(5,5), c(5,5), r(5)
!HPF$    sequence lhs, c, r
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine binvrhs( lhs,r )
         double precision lhs(5,5), r(5)
!HPF$    sequence lhs, r
         end subroutine
      end interface

      isize = grid_points(1)-1

c---------------------------------------------------------------------
c     outer most do loops - sweeping in i direction
c---------------------------------------------------------------------
!HPF$ independent
      do k=1,grid_points(3)-2 

         do j=1,grid_points(2)-2
            do jj=1, 5
               do ii=1, 5
                  cblock(ii,jj,j) = lhs(ii,jj,bb,0,j,k)
                  dblock(ii,jj,j) = lhs(ii,jj,cc,0,j,k)
               end do
               bvec(jj,j) = rhs(jj,0,j,k)
            end do
         end do

         do j=1,grid_points(2)-2

c---------------------------------------------------------------------
c     multiply c(0,j,k) by b_inverse and copy back to c
c     multiply rhs(0) by b_inverse(0) and copy to rhs
c---------------------------------------------------------------------
            call binvcrhs(cblock(1,1,j), dblock(1,1,j), bvec(1,j))

         end do

         do j=1,grid_points(2)-2
            do jj=1, 5
               do ii=1, 5
                  lhs(ii,jj,bb,0,j,k)   = cblock(ii,jj,j)
                  lhs(ii,jj,cc,0,j,k)   = dblock(ii,jj,j)
               end do
               rhs(jj,0,j,k)   = bvec(jj,j)
            end do
         enddo

      enddo

c---------------------------------------------------------------------
c     begin inner most do loop
c     do all the elements of the cell unless last 
c---------------------------------------------------------------------
!HPF$ independent
      do k=1,grid_points(3)-2 
         do i=1,isize-1

            do j=1,grid_points(2)-2
               do jj=1, 5
                  do ii=1, 5
                     ablock(ii,jj,j) = lhs(ii,jj,aa,i,j,k)
                     bblock(ii,jj,j) = lhs(ii,jj,cc,i-1,j,k)
                     cblock(ii,jj,j) = lhs(ii,jj,bb,i,j,k)
                     dblock(ii,jj,j) = lhs(ii,jj,cc,i,j,k)
                  end do
                  avec(jj,j) = rhs(jj,i-1,j,k)
                  bvec(jj,j) = rhs(jj,i,j,k)
               end do
            end do

            do j=1,grid_points(2)-2

c---------------------------------------------------------------------
c     rhs(i) = rhs(i) - A*rhs(i-1)
c---------------------------------------------------------------------
               call matvec_sub(ablock(1,1,j), avec(1,j), bvec(1,j))

c---------------------------------------------------------------------
c     B(i) = B(i) - C(i-1)*A(i)
c---------------------------------------------------------------------
               call matmul_sub(ablock(1,1,j), bblock(1,1,j),
     >                         cblock(1,1,j))

c---------------------------------------------------------------------
c     multiply c(i,j,k) by b_inverse and copy back to c
c     multiply rhs(1,j,k) by b_inverse(1,j,k) and copy to rhs
c---------------------------------------------------------------------
               call binvcrhs(cblock(1,1,j), dblock(1,1,j), bvec(1,j))

            end do

            do j=1,grid_points(2)-2
               do jj=1, 5
                  do ii=1, 5
                     lhs(ii,jj,aa,i,j,k)   = ablock(ii,jj,j)
                     lhs(ii,jj,cc,i-1,j,k) = bblock(ii,jj,j)
                     lhs(ii,jj,bb,i,j,k)   = cblock(ii,jj,j)
                     lhs(ii,jj,cc,i,j,k)   = dblock(ii,jj,j)
                  end do
                  rhs(jj,i-1,j,k) = avec(jj,j)
                  rhs(jj,i,j,k)   = bvec(jj,j)
               end do
            enddo

         enddo
      enddo

!HPF$ independent
      do k=1,grid_points(3)-2

         do j=1,grid_points(2)-2
            do jj=1, 5
               do ii=1, 5
                  ablock(ii,jj,j) = lhs(ii,jj,aa,isize,j,k)
                  bblock(ii,jj,j) = lhs(ii,jj,cc,isize-1,j,k)
                  cblock(ii,jj,j) = lhs(ii,jj,bb,isize,j,k)
               end do
               avec(jj,j) = rhs(jj,isize-1,j,k)
               bvec(jj,j) = rhs(jj,isize,j,k)
            end do
         end do

         do j=1,grid_points(2)-2

c---------------------------------------------------------------------
c     rhs(isize) = rhs(isize) - A*rhs(isize-1)
c---------------------------------------------------------------------
            call matvec_sub(ablock(1,1,j), avec(1,j), bvec(1,j))

c---------------------------------------------------------------------
c     B(isize) = B(isize) - C(isize-1)*A(isize)
c---------------------------------------------------------------------
            call matmul_sub(ablock(1,1,j), bblock(1,1,j),
     >                      cblock(1,1,j))

c---------------------------------------------------------------------
c     multiply rhs() by b_inverse() and copy to rhs
c---------------------------------------------------------------------
            call binvrhs(cblock(1,1,j), bvec(1,j))

         enddo

         do j=1,grid_points(2)-2
            do jj=1, 5
               do ii=1, 5
                  lhs(ii,jj,aa,isize,j,k)   = ablock(ii,jj,j)
                  lhs(ii,jj,cc,isize-1,j,k) = bblock(ii,jj,j)
                  lhs(ii,jj,bb,isize,j,k)   = cblock(ii,jj,j)
               end do
               rhs(jj,isize-1,j,k) = avec(jj,j)
               rhs(jj,isize,j,k)   = bvec(jj,j)
            end do
         enddo

      enddo

      return
      end
