c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine y_solve

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     Performs line solves in Y direction by first factoring
c     the block-tridiagonal matrix into an upper triangular matrix, 
c     and then performing back substitution to solve for the unknow
c     vectors of each line.  
c     
c     Make sure we treat elements zero to cell_size in the direction
c     of the sweep.
c---------------------------------------------------------------------

      include 'header.h'

      call lhsy
      call y_solve_cell
      call y_backsubstitute

      return
      end
      

      subroutine y_backsubstitute

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     back solve: if last cell, then generate U(jsize)=rhs(jsize)
c     else assume U(jsize) is loaded in un pack backsub_info
c     so just use it
c     after call u(jstart) will be sent to next cell
c---------------------------------------------------------------------

      include 'header.h'

      integer i, j, k, m,n
      
      do k=1,grid_points(3)-2
         do j=grid_points(2)-2,0,-1
            do i=1,grid_points(1)-2
               do m=1,BLOCK_SIZE
                  do n=1,BLOCK_SIZE
                     rhs(m,i,j,k) = rhs(m,i,j,k) 
     >                    - lhs(m,n,cc,i,j,k)*rhs(n,i,j+1,k)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine y_solve_cell

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     performs guaussian elimination on this cell.
c     
c     assumes that unpacking routines for non-first cells 
c     preload C' and rhs' from previous cell.
c     
c     assumed send happens outside this routine, but that
c     c'(JMAX) and rhs'(JMAX) will be sent to next cell
c---------------------------------------------------------------------

      include 'header.h'

      integer i,j,k,jsize,jstart

      double precision ablock(5,5,0:IMAX/2*2), bblock(5,5,0:IMAX/2*2),
     >                 cblock(5,5,0:IMAX/2*2), dblock(5,5,0:IMAX/2*2)
      double precision avec(5,0:IMAX/2*2), bvec(5,0:IMAX/2*2)
!HPF$ sequence ablock, bblock, cblock, dblock, avec, bvec

      integer ii, jj

      interface
         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine matvec_sub(ablock,avec,bvec)
         double precision ablock,avec,bvec
         dimension ablock(5,5),avec(5),bvec(5)
!HPF$    sequence ablock, avec, bvec
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine matmul_sub(ablock, bblock, cblock)
         double precision ablock, bblock, cblock
         dimension ablock(5,5), bblock(5,5), cblock(5,5)
!HPF$    sequence ablock, bblock, cblock
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine binvcrhs( lhs,c,r )
         double precision lhs(5,5), c(5,5), r(5)
!HPF$    sequence lhs, c, r
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine binvrhs( lhs,r )
         double precision lhs(5,5), r(5)
!HPF$    sequence lhs, r
         end subroutine
      end interface

      jstart = 0
      jsize = grid_points(2)-1

!HPF$ independent
      do k=1,grid_points(3)-2 

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  cblock(ii,jj,i) = lhs(ii,jj,bb,i,0,k)
                  dblock(ii,jj,i) = lhs(ii,jj,cc,i,0,k)
               end do
               bvec(jj,i) = rhs(jj,i,0,k)
            end do
         end do

         do i=1,grid_points(1)-2

c---------------------------------------------------------------------
c     multiply c(i,0,k) by b_inverse and copy back to c
c     multiply rhs(0) by b_inverse(0) and copy to rhs
c---------------------------------------------------------------------
            call binvcrhs(cblock(1,1,i), dblock(1,1,i), bvec(1,i))
         enddo

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  lhs(ii,jj,bb,i,0,k)   = cblock(ii,jj,i)
                  lhs(ii,jj,cc,i,0,k)   = dblock(ii,jj,i)
               end do
               rhs(jj,i,0,k)   = bvec(jj,i)
            end do
         end do

      enddo

c---------------------------------------------------------------------
c     begin inner most do loop
c     do all the elements of the cell unless last 
c---------------------------------------------------------------------
!HPF$ independent
      do k=1,grid_points(3)-2 
         do j=1,jsize-1

            do i=1,grid_points(1)-2
               do jj=1, 5
                  do ii=1, 5
                     ablock(ii,jj,i) = lhs(ii,jj,aa,i,j,k)
                     bblock(ii,jj,i) = lhs(ii,jj,cc,i,j-1,k)
                     cblock(ii,jj,i) = lhs(ii,jj,bb,i,j,k)
                     dblock(ii,jj,i) = lhs(ii,jj,cc,i,j,k)
                  end do
                  avec(jj,i) = rhs(jj,i,j-1,k)
                  bvec(jj,i) = rhs(jj,i,j,k)
               end do
            end do

            do i=1,grid_points(1)-2

c---------------------------------------------------------------------
c     subtract A*lhs_vector(j-1) from lhs_vector(j)
c     
c     rhs(j) = rhs(j) - A*rhs(j-1)
c---------------------------------------------------------------------
               call matvec_sub(ablock(1,1,i), avec(1,i),
     >                         bvec(1,i))

c---------------------------------------------------------------------
c     B(j) = B(j) - C(j-1)*A(j)
c---------------------------------------------------------------------
               call matmul_sub(ablock(1,1,i), bblock(1,1,i),
     >                         cblock(1,1,i))

c---------------------------------------------------------------------
c     multiply c(i,j,k) by b_inverse and copy back to c
c     multiply rhs(i,1,k) by b_inverse(i,1,k) and copy to rhs
c---------------------------------------------------------------------
               call binvcrhs(cblock(1,1,i), dblock(1,1,i), bvec(1,i))

            enddo

            do i=1,grid_points(1)-2
               do jj=1, 5
                  do ii=1, 5
                     lhs(ii,jj,aa,i,j,k)   = ablock(ii,jj,i)
                     lhs(ii,jj,cc,i,j-1,k) = bblock(ii,jj,i)
                     lhs(ii,jj,bb,i,j,k)   = cblock(ii,jj,i)
                     lhs(ii,jj,cc,i,j,k)   = dblock(ii,jj,i)
                  end do
                  rhs(jj,i,j-1,k) = avec(jj,i)
                  rhs(jj,i,j,k)   = bvec(jj,i)
               end do
            end do

         enddo
      enddo

!HPF$ independent
      do k=1,grid_points(3)-2 

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  ablock(ii,jj,i) = lhs(ii,jj,aa,i,jsize,k)
                  bblock(ii,jj,i) = lhs(ii,jj,cc,i,jsize-1,k)
                  cblock(ii,jj,i) = lhs(ii,jj,bb,i,jsize,k)
               end do
               avec(jj,i) = rhs(jj,i,jsize-1,k)
               bvec(jj,i) = rhs(jj,i,jsize,k)
            end do
         end do

         do i=1,grid_points(1)-2

c---------------------------------------------------------------------
c     rhs(jsize) = rhs(jsize) - A*rhs(jsize-1)
c---------------------------------------------------------------------
            call matvec_sub(ablock(1,1,i), avec(1,i), bvec(1,i))

c---------------------------------------------------------------------
c     B(jsize) = B(jsize) - C(jsize-1)*A(jsize)
c     call matmul_sub(aa,i,jsize,k,c,
c     $              cc,i,jsize-1,k,c,bb,i,jsize,k)
c---------------------------------------------------------------------
            call matmul_sub(ablock(1,1,i), bblock(1,1,i),
     >                      cblock(1,1,i))

c---------------------------------------------------------------------
c     multiply rhs(jsize) by b_inverse(jsize) and copy to rhs
c---------------------------------------------------------------------
            call binvrhs(cblock(1,1,i), bvec(1,i))

         enddo

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  lhs(ii,jj,aa,i,jsize,k)   = ablock(ii,jj,i)
                  lhs(ii,jj,cc,i,jsize-1,k) = bblock(ii,jj,i)
                  lhs(ii,jj,bb,i,jsize,k)   = cblock(ii,jj,i)
               end do
               rhs(jj,i,jsize-1,k) = avec(jj,i)
               rhs(jj,i,jsize,k)   = bvec(jj,i)
            end do
         end do

      enddo

      return
      end
      


