c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine z_solve

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     Performs line solves in Z direction by first factoring
c     the block-tridiagonal matrix into an upper triangular matrix, 
c     and then performing back substitution to solve for the unknow
c     vectors of each line.  
c     
c     Make sure we treat elements zero to cell_size in the direction
c     of the sweep.
c---------------------------------------------------------------------

      include 'header.h'

      double precision rhs2(5,     0:IMAX/2*2, 0:JMAX/2*2, 0:KMAX/2*2),
     >                 lhs2(5,5,3, 0:IMAX/2*2, 0:JMAX/2*2, 0:KMAX/2*2)
!HPF$ align (*,*,i,*) with ttt(i) :: rhs2
!HPF$ align (*,*,*,*,i,*) with ttt(i) :: lhs2

      call lhsz

      rhs2 = rhs
      lhs2 = lhs

      call z_solve_cell(rhs2, lhs2)
      call z_backsubstitute(rhs2, lhs2)

      rhs = rhs2
      lhs = lhs2

      return
      end
      
c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine z_backsubstitute(rhs2, lhs2)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     back solve: if last cell, then generate U(ksize)=rhs(ksize)
c     else assume U(ksize) is loaded in un pack backsub_info
c     so just use it
c     after call u(kstart) will be sent to next cell
c---------------------------------------------------------------------

      include 'header.h'

      integer i, j, k, m, n
      
      double precision rhs2(5,     0:IMAX/2*2, 0:JMAX/2*2, 0:KMAX/2*2),
     >                 lhs2(5,5,3, 0:IMAX/2*2, 0:JMAX/2*2, 0:KMAX/2*2)
!HPF$ align (*,*,i,*) with *ttt(i) :: rhs2
!HPF$ align (*,*,*,*,i,*) with *ttt(i) :: lhs2

      do k=grid_points(3)-2,0,-1
         do j=1,grid_points(2)-2
            do i=1,grid_points(1)-2
               do m=1,BLOCK_SIZE
                  do n=1,BLOCK_SIZE
                     rhs2(m,i,j,k) = rhs2(m,i,j,k) 
     >                    - lhs2(m,n,cc,i,j,k)*rhs2(n,i,j,k+1)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine z_solve_cell(rhs2, lhs2)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     performs guaussian elimination on this cell.
c     
c     assumes that unpacking routines for non-first cells 
c     preload C' and rhs' from previous cell.
c     
c     assumed send happens outside this routine, but that
c     c'(KMAX) and rhs'(KMAX) will be sent to next cell.
c---------------------------------------------------------------------

      include 'header.h'

      integer i,j,k,ksize

      double precision rhs2(5,     0:IMAX/2*2, 0:JMAX/2*2, 0:KMAX/2*2),
     >                 lhs2(5,5,3, 0:IMAX/2*2, 0:JMAX/2*2, 0:KMAX/2*2)
!HPF$ align (*,*,i,*) with *ttt(i) :: rhs2
!HPF$ align (*,*,*,*,i,*) with *ttt(i) :: lhs2

      double precision ablock(5,5,0:IMAX/2*2), bblock(5,5,0:IMAX/2*2),
     >                 cblock(5,5,0:IMAX/2*2), dblock(5,5,0:IMAX/2*2)
      double precision avec(5,0:IMAX/2*2), bvec(5,0:IMAX/2*2)
!HPF$ sequence ablock, bblock, cblock, dblock, avec, bvec

      integer ii, jj

      interface
         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine matvec_sub(ablock,avec,bvec)
         double precision ablock,avec,bvec
         dimension ablock(5,5),avec(5),bvec(5)
!HPF$    sequence ablock, avec, bvec
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine matmul_sub(ablock, bblock, cblock)
         double precision ablock, bblock, cblock
         dimension ablock(5,5), bblock(5,5), cblock(5,5)
!HPF$    sequence ablock, bblock, cblock
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine binvcrhs( lhs,c,r )
         double precision lhs(5,5), c(5,5), r(5)
!HPF$    sequence lhs, c, r
         end subroutine

         pure extrinsic (FORTRAN_LOCAL)
     >     subroutine binvrhs( lhs,r )
         double precision lhs(5,5), r(5)
!HPF$    sequence lhs, r
         end subroutine
      end interface

      ksize = grid_points(3)-1

c---------------------------------------------------------------------
c     outer most do loops - sweeping in i direction
c---------------------------------------------------------------------
!HPF$ independent
      do j=1,grid_points(2)-2 

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  cblock(ii,jj,i) = lhs2(ii,jj,bb,i,j,0)
                  dblock(ii,jj,i) = lhs2(ii,jj,cc,i,j,0)
               end do
               bvec(jj,i) = rhs2(jj,i,j,0)
            end do
         end do

         do i=1,grid_points(1)-2

c---------------------------------------------------------------------
c     multiply c(i,j,0) by b_inverse and copy back to c
c     multiply rhs(0) by b_inverse(0) and copy to rhs
c---------------------------------------------------------------------
            call binvcrhs(cblock(1,1,i), dblock(1,1,i), bvec(1,i))

         enddo

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  lhs2(ii,jj,bb,i,j,0)   = cblock(ii,jj,i)
                  lhs2(ii,jj,cc,i,j,0)   = dblock(ii,jj,i)
               end do
               rhs2(jj,i,j,0)   = bvec(jj,i)
            end do
         enddo

      enddo

c---------------------------------------------------------------------
c     begin inner most do loop
c     do all the elements of the cell unless last 
c---------------------------------------------------------------------
!HPF$ independent
      do j=1,grid_points(2)-2
         do k=1,ksize-1

            do i=1,grid_points(1)-2
               do jj=1, 5
                  do ii=1, 5
                     ablock(ii,jj,i) = lhs2(ii,jj,aa,i,j,k)
                     bblock(ii,jj,i) = lhs2(ii,jj,cc,i,j,k-1)
                     cblock(ii,jj,i) = lhs2(ii,jj,bb,i,j,k)
                     dblock(ii,jj,i) = lhs2(ii,jj,cc,i,j,k)
                  end do
                  avec(jj,i) = rhs2(jj,i,j,k-1)
                  bvec(jj,i) = rhs2(jj,i,j,k)
               end do
            end do

            do i=1,grid_points(1)-2

c---------------------------------------------------------------------
c     subtract A*lhs_vector(k-1) from lhs_vector(k)
c     
c     rhs(k) = rhs(k) - A*rhs(k-1)
c---------------------------------------------------------------------
               call matvec_sub(ablock(1,1,i), avec(1,i), bvec(1,i))

c---------------------------------------------------------------------
c     B(k) = B(k) - C(k-1)*A(k)
c     call matmul_sub(aa,i,j,k,c,cc,i,j,k-1,c,bb,i,j,k)
c---------------------------------------------------------------------
               call matmul_sub(ablock(1,1,i), bblock(1,1,i),
     >                         cblock(1,1,i))

c---------------------------------------------------------------------
c     multiply c(i,j,k) by b_inverse and copy back to c
c     multiply rhs(i,j,1) by b_inverse(i,j,1) and copy to rhs
c---------------------------------------------------------------------
               call binvcrhs(cblock(1,1,i), dblock(1,1,i), bvec(1,i))

            end do

            do i=1,grid_points(1)-2
               do jj=1, 5
                  do ii=1, 5
                     lhs2(ii,jj,aa,i,j,k)   = ablock(ii,jj,i)
                     lhs2(ii,jj,cc,i,j,k-1) = bblock(ii,jj,i)
                     lhs2(ii,jj,bb,i,j,k)   = cblock(ii,jj,i)
                     lhs2(ii,jj,cc,i,j,k)   = dblock(ii,jj,i)
                  end do
                  rhs2(jj,i,j,k-1) = avec(jj,i)
                  rhs2(jj,i,j,k)   = bvec(jj,i)
               end do
            enddo

         enddo
      enddo

c---------------------------------------------------------------------
c     Now finish up special cases for last cell
c---------------------------------------------------------------------
!HPF$ independent
      do j=1,grid_points(2)-2

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  ablock(ii,jj,i) = lhs2(ii,jj,aa,i,j,ksize)
                  bblock(ii,jj,i) = lhs2(ii,jj,cc,i,j,ksize-1)
                  cblock(ii,jj,i) = lhs2(ii,jj,bb,i,j,ksize)
               end do
               avec(jj,i) = rhs2(jj,i,j,ksize-1)
               bvec(jj,i) = rhs2(jj,i,j,ksize)
            end do
         end do

         do i=1,grid_points(1)-2

c---------------------------------------------------------------------
c     rhs(ksize) = rhs(ksize) - A*rhs(ksize-1)
c---------------------------------------------------------------------
            call matvec_sub(ablock(1,1,i), avec(1,i), bvec(1,i))

c---------------------------------------------------------------------
c     B(ksize) = B(ksize) - C(ksize-1)*A(ksize)
c     call matmul_sub(aa,i,j,ksize,c,
c     $              cc,i,j,ksize-1,c,bb,i,j,ksize)
c---------------------------------------------------------------------
            call matmul_sub(ablock(1,1,i), bblock(1,1,i),
     >                      cblock(1,1,i))

c---------------------------------------------------------------------
c     multiply rhs(ksize) by b_inverse(ksize) and copy to rhs
c---------------------------------------------------------------------
            call binvrhs(cblock(1,1,i), bvec(1,i))

         enddo

         do i=1,grid_points(1)-2
            do jj=1, 5
               do ii=1, 5
                  lhs2(ii,jj,aa,i,j,ksize)   = ablock(ii,jj,i)
                  lhs2(ii,jj,cc,i,j,ksize-1) = bblock(ii,jj,i)
                  lhs2(ii,jj,bb,i,j,ksize)   = cblock(ii,jj,i)
               end do
               rhs2(jj,i,j,ksize-1) = avec(jj,i)
               rhs2(jj,i,j,ksize)   = bvec(jj,i)
            end do
         enddo

      enddo

      return
      end
      





