!---------------------------------------------------------------------
!   NAS Parallel Benchmark - MG
!
!   HPF/JA Version derived from NPB2.3-serial
!
!    coded by H. Murai (Earth Simulator Center)
!---------------------------------------------------------------------

      module mapping
      integer, allocatable :: map(:)
      end module

      program mg_main

      use mapping

      include 'globals.h'

      integer np, lsz

      np = number_of_processors()

      allocate( map(np) )

      lsz = nz_default / np

      if (np == 1) then
         map(1) = lsz + 2 + 2**(lt_default-1)
      else
         do i=1, np-1
            map(i) = lsz
         end do
         map(np) = lsz + 2 + 2**(lt_default-1)
      end if

      call mg

      end

c---------------------------------------------------------------------
      subroutine mg
c---------------------------------------------------------------------

      use mapping

      implicit none

      include 'globals.h'

c---------------------------------------------------------------------------c
c k is the current level. It is passed down through subroutine args
c and is NOT global. it is the current iteration
c---------------------------------------------------------------------------c

      integer k, it
      
      external timer_read
      double precision t, tinit, mflops, timer_read

c---------------------------------------------------------------------------c
c These arrays are in common because they are quite large
c and probably shouldn't be allocated on the stack. They
c are always passed as subroutine args. 
c---------------------------------------------------------------------------c

      double precision,
     > dimension(nx_default+2, ny_default+2, nz_default+2) :: u, v, r
!HPF$ template ttt(nz_default+2+2**(lt_default-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with ttt(i) :: u, v, r
!HPF$ shadow (0,0,1:1) :: u, v, r
      double precision a(0:3),c(0:3)

      double precision rnm2, rnmu, old2, oldu, epsilon
      integer n1, n2, n3, nn, nit
      double precision verify_value
      logical verified

      integer ierr,i, fstatus, nprocs
      integer T_bench, T_init
      parameter (T_bench=1, T_init=2)
      integer i1,i2,i3
      call timer_clear(T_bench)
      call timer_clear(T_init)


      call timer_start(T_init)
      

c---------------------------------------------------------------------
c Read in and broadcast input data
c---------------------------------------------------------------------

      write (*, 1000) 

      open(unit=7,file="mg.input", status="old", iostat=fstatus)
      if (fstatus .eq. 0) then
         write(*,50) 
 50      format(' Reading from input file mg.input')
         read(7,*) lt
         read(7,*) nx(lt), ny(lt), nz(lt)
         read(7,*) nit
         read(7,*) (debug_vec(i),i=0,7)
      else
         write(*,51) 
 51      format(' No input file. Using compiled defaults ')
         lt = lt_default
         nit = nit_default
         nx(lt) = nx_default
         ny(lt) = ny_default
         nz(lt) = nz_default
         do i = 0,7
            debug_vec(i) = debug_default
         end do
      endif

      if ( (nx(lt) .ne. ny(lt)) .or. (nx(lt) .ne. nz(lt)) ) then
         Class = 'U' 
      else if( nx(lt) .eq. 32 .and. nit .eq. 4 ) then
         Class = 'S'
      else if( nx(lt) .eq. 64 .and. nit .eq. 40 ) then
         Class = 'W'
      else if( nx(lt) .eq. 256 .and. nit .eq. 20 ) then
         Class = 'B'
      else if( nx(lt) .eq. 512 .and. nit .eq. 20 ) then  
         Class = 'C'
      else if( nx(lt) .eq. 256 .and. nit .eq. 4 ) then  
         Class = 'A'
      else
         Class = 'U'
      endif

c---------------------------------------------------------------------
c  Use these for debug info:
c---------------------------------------------------------------------
c     debug_vec(0) = 1 !=> report all norms
c     debug_vec(1) = 1 !=> some setup information
c     debug_vec(1) = 2 !=> more setup information
c     debug_vec(2) = k => at level k or below, show result of resid
c     debug_vec(3) = k => at level k or below, show result of psinv
c     debug_vec(4) = k => at level k or below, show result of rprj
c     debug_vec(5) = k => at level k or below, show result of interp
c     debug_vec(6) = 1 => (unused)
c     debug_vec(7) = 1 => (unused)
c---------------------------------------------------------------------
      a(0) = -8.0D0/3.0D0 
      a(1) =  0.0D0 
      a(2) =  1.0D0/6.0D0 
      a(3) =  1.0D0/12.0D0
      
      if(Class .eq. 'A' .or. Class .eq. 'S'.or. Class .eq.'W') then
c---------------------------------------------------------------------
c     Coefficients for the S(a) smoother
c---------------------------------------------------------------------
         c(0) =  -3.0D0/8.0D0
         c(1) =  +1.0D0/32.0D0
         c(2) =  -1.0D0/64.0D0
         c(3) =   0.0D0
      else
c---------------------------------------------------------------------
c     Coefficients for the S(b) smoother
c---------------------------------------------------------------------
         c(0) =  -3.0D0/17.0D0
         c(1) =  +1.0D0/33.0D0
         c(2) =  -1.0D0/61.0D0
         c(3) =   0.0D0
      endif
      lb = 1
      k  = lt

      call setup(n1,n2,n3,k)
      call zero3(u,n1,n2,n3,k)
      call zran3(v,n1,n2,n3,nx(lt),ny(lt),k)

      call norm2u3(v,n1,n2,n3,rnm2,rnmu,nx(lt),ny(lt),nz(lt),k)
c     write(*,*)
c     write(*,*)' norms of random v are'
c     write(*,600) 0, rnm2, rnmu
c     write(*,*)' about to evaluate resid, k=',k

      write (*, 1001) nx(lt),ny(lt),nz(lt), Class
      write (*, 1002) nit
      nprocs = number_of_processors()
      write (*, 1003) nprocs

 1000 format(//,' NAS Parallel Benchmarks HPF/JA version',
     >          ' - MG Benchmark', /)
 1001 format(' Size: ', i3, 'x', i3, 'x', i3, '  (class ', A, ')' )
 1002 format(' Iterations: ', i3)
 1003 format(' Number of active processes: ', i5, /)

      call resid(u,v,r,n1,n2,n3,a,k)
      call norm2u3(r,n1,n2,n3,rnm2,rnmu,nx(lt),ny(lt),nz(lt),k)
      old2 = rnm2
      oldu = rnmu

c---------------------------------------------------------------------
c     One iteration for startup
c---------------------------------------------------------------------
      call mg3P(u,v,r,a,c,n1,n2,n3,k)
      call resid(u,v,r,n1,n2,n3,a,k)
      call setup(n1,n2,n3,k)
      call zero3(u,n1,n2,n3,k)
      call zran3(v,n1,n2,n3,nx(lt),ny(lt),k)

      call timer_stop(T_init)
      call timer_start(T_bench)

      call resid(u,v,r,n1,n2,n3,a,k)
      call norm2u3(r,n1,n2,n3,rnm2,rnmu,nx(lt),ny(lt),nz(lt),k)
      old2 = rnm2
      oldu = rnmu

      do  it=1,nit
         call mg3P(u,v,r,a,c,n1,n2,n3,k)
         call resid(u,v,r,n1,n2,n3,a,k)
      enddo

      call norm2u3(r,n1,n2,n3,rnm2,rnmu,nx(lt),ny(lt),nz(lt),k)

      call timer_stop(T_bench)

      t = timer_read(T_bench)
      tinit = timer_read(T_init)

      verified = .FALSE.
      verify_value = 0.0

      write( *,'(/A,F15.3,A/)' ) 
     >     ' Initialization time: ',tinit, ' seconds'
      write(*,100)
 100  format(' Benchmark completed ')

      epsilon = 1.d-8
      if (Class .ne. 'U') then
         if(Class.eq.'S') then
            verify_value = 0.530770700573d-04
         elseif(Class.eq.'W') then
            verify_value = 0.250391406439E-17 ! 40 iterations
!				0.183103168997d-044 iterations
         elseif(Class.eq.'A') then
            verify_value = 0.2433365309d-5
         elseif(Class.eq.'B') then
            verify_value = 0.180056440132d-5
         elseif(Class.eq.'C') then
            verify_value = 0.570674826298d-06
         endif

         if( abs( rnm2 - verify_value ) .le. epsilon ) then
            verified = .TRUE.
            write(*, 200)
            write(*, 201) rnm2
            write(*, 202) rnm2 - verify_value
 200        format(' VERIFICATION SUCCESSFUL ')
 201        format(' L2 Norm is ', E20.12)
 202        format(' Error is   ', E20.12)
         else
            verified = .FALSE.
            write(*, 300) 
            write(*, 301) rnm2
            write(*, 302) verify_value
 300        format(' VERIFICATION FAILED')
 301        format(' L2 Norm is             ', E20.12)
 302        format(' The correct L2 Norm is ', E20.12)
         endif
      else
         verified = .FALSE.
         write (*, 400)
         write (*, 401)
 400     format(' Problem size unknown')
 401     format(' NO VERIFICATION PERFORMED')
      endif

      nn = nx(lt)*ny(lt)*nz(lt)

      if( t .ne. 0. ) then
         mflops = 58.*nit*nn*1.0D-6 /t
      else
         mflops = 0.0
      endif

      call print_results('MG', class, nx(lt), ny(lt), nz(lt), 
     >                   nit, nprocs, t,
     >                   mflops, '          floating point', 
     >                   verified, npbversion, compiletime,
     >                   cs1, cs2, cs3, cs4, cs5, cs6, cs7)


 600  format( i4, 2e19.12)

      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine setup(n1,n2,n3,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      use mapping

      implicit none

      include 'globals.h'

      integer  is1, is2, is3, ie1, ie2, ie3
      common /grid/ is1,is2,is3,ie1,ie2,ie3

      integer n1,n2,n3,k
      integer d, i, j

      integer ax, mi(3,10)
      integer ng(3,10)
      integer s, dir,ierr


      ng(1,lt) = nx(lt)
      ng(2,lt) = ny(lt)
      ng(3,lt) = nz(lt)
      do  ax=1,3
         do  k=lt-1,1,-1
            ng(ax,k) = ng(ax,k+1)/2
         enddo
      enddo
 61   format(10i4)
      do  k=lt,1,-1
         nx(k) = ng(1,k)
         ny(k) = ng(2,k)
         nz(k) = ng(3,k)
      enddo

      do  k = lt,1,-1
         do  ax = 1,3
            mi(ax,k) = 2 + ng(ax,k) 
         enddo

         m1(k) = mi(1,k)
         m2(k) = mi(2,k)
         m3(k) = mi(3,k)

      enddo

      k = lt
      is1 = 2 + ng(1,k) - ng(1,lt)
      ie1 = 1 + ng(1,k)
      n1 = 3 + ie1 - is1
      is2 = 2 + ng(2,k) - ng(2,lt)
      ie2 = 1 + ng(2,k) 
      n2 = 3 + ie2 - is2
      is3 = 2 + ng(3,k) - ng(3,lt)
      ie3 = 1 + ng(3,k) 
      n3 = 3 + ie3 - is3

      if( debug_vec(1) .ge. 1 )then
         write(*,*)' in setup, '
         write(*,*)' k  lt  nx  ny  nz ',
     >        ' n1  n2  n3 is1 is2 is3 ie1 ie2 ie3'
         write(*,9) k,lt,ng(1,k),ng(2,k),ng(3,k),
     >              n1,n2,n3,is1,is2,is3,ie1,ie2,ie3
 9       format(15i4)
      endif

      k = lt

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      recursive subroutine mg3P(u,v,r,a,c,n1,n2,n3,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     multigrid V-cycle routine
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'globals.h'

      integer n1, n2, n3, k
      double precision u(n1,n2,n3), v(:,:,:), r(n1,n2,n3)
!HPF$ template ttt(nz(lt)+2+2**(lt-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with *ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: u, r
!HPF$ align (*,*,i) with *ttt(i) :: v
!HPF$ shadow (0,0,1:1) :: u, v, r
      double precision a(0:3),c(0:3)

      integer j

      double precision, allocatable :: u2(:,:,:), r2(:,:,:)
!HPF$ align (*,*,i) with ttt(i*(2**(lt+1-k))-(2**(lt+1-k)-1)) :: u2, r2
!HPF$ shadow (0,0,1:1) :: u2, r2
c---------------------------------------------------------------------
c     down cycle.
c     restrict the residual from the find grid to the coarse
c---------------------------------------------------------------------
      j = k-1

      allocate(u2(m1(j), m2(j), m3(j)))
      allocate(r2(m1(j), m2(j), m3(j)))

      call rprj3(r,m1(k),m2(k),m3(k), r2,m1(j),m2(j),m3(j), k)

      if (k > lb+1) then
         call mg3p(u2,v,r2,a,c,m1(j),m2(j),m3(j),j)
      end if

      if (k == lb+1) then

c---------------------------------------------------------------------
c        compute an approximate solution on the coarsest grid
c---------------------------------------------------------------------
         call zero3(u2,m1(j),m2(j),m3(j),j)
         call psinv(r2,u2,m1(j),m2(j),m3(j),c,j)

      end if

      if (k < lt) then

c---------------------------------------------------------------------
c        prolongate from level k-1  to k
c---------------------------------------------------------------------
         call zero3(u,m1(k),m2(k),m3(k),k)
         call interp(u2,m1(j),m2(j),m3(j), u,m1(k),m2(k),m3(k), k)

c---------------------------------------------------------------------
c        compute residual for level k
c---------------------------------------------------------------------
         call resid(u,r,r,m1(k),m2(k),m3(k),a,k)

c---------------------------------------------------------------------
c        apply smoother
c---------------------------------------------------------------------
         call psinv(r,u,m1(k),m2(k),m3(k),c,k)

      else if (k == lt) then

         call interp(u2,m1(j),m2(j),m3(j), u,n1,n2,n3, k)
         call resid(u,v,r,n1,n2,n3,a,k)
         call psinv(r,u,n1,n2,n3,c,k)

      end if

      deallocate(u2)
      deallocate(r2)

      return

      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine psinv(r,u,n1,n2,n3,c,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     psinv applies an approximate inverse as smoother:  u = u + Cr
c
c     This  implementation costs  15A + 4M per result, where
c     A and M denote the costs of Addition and Multiplication.  
c     Presuming coefficient c(3) is zero (the NPB assumes this,
c     but it is thus not a general case), 2A + 1M may be eliminated,
c     resulting in 13A + 3M.
c     Note that this vectorizes, and is also fine for cache 
c     based machines.  
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'globals.h'

      integer n1,n2,n3,k
      double precision u(n1,n2,n3),r(n1,n2,n3),c(0:3)
!HPF$ template ttt(nz(lt)+2+2**(lt-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with *ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: u, r
!HPF$ shadow (0,0,1:1) :: u, r
      integer i3, i2, i1

      double precision r1(m), r2(m)

!HPFJ reflect r

!HPF$ independent
      do i3=2,n3-1
!HPF$    on home(u(:,:,i3)), local begin
         do i2=2,n2-1
            do i1=1,n1
               r1(i1) = r(i1,i2-1,i3) + r(i1,i2+1,i3)
     >                + r(i1,i2,i3-1) + r(i1,i2,i3+1)
               r2(i1) = r(i1,i2-1,i3-1) + r(i1,i2+1,i3-1)
     >                + r(i1,i2-1,i3+1) + r(i1,i2+1,i3+1)
            enddo
            do i1=2,n1-1
               u(i1,i2,i3) = u(i1,i2,i3)
     >                     + c(0) * r(i1,i2,i3)
     >                     + c(1) * ( r(i1-1,i2,i3) + r(i1+1,i2,i3)
     >                              + r1(i1) )
     >                     + c(2) * ( r2(i1) + r1(i1-1) + r1(i1+1) )
            enddo
         enddo
!HPF$    end on
      enddo

c---------------------------------------------------------------------
c     exchange boundary points
c---------------------------------------------------------------------
      u(n1,:,:) = u(2,:,:)
      u(1,:,:) = u(n1-1,:,:)
      u(:,n2,:) = u(:,2,:)
      u(:,1,:) = u(:,n2-1,:)       
      u(:,:,n3) = u(:,:,2)
      u(:,:,1) = u(:,:,n3-1)

      if( debug_vec(0) .ge. 1 )then
         call rep_nrm(u,n1,n2,n3,'   psinv',k)
      endif

      if( debug_vec(3) .ge. k )then
         call showall(u,n1,n2,n3,k)
      endif

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine resid(u,v,r,n1,n2,n3,a,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     resid computes the residual:  r = v - Au
c
c     This  implementation costs  15A + 4M per result, where
c     A and M denote the costs of Addition (or Subtraction) and 
c     Multiplication, respectively. 
c     Presuming coefficient a(1) is zero (the NPB assumes this,
c     but it is thus not a general case), 3A + 1M may be eliminated,
c     resulting in 12A + 3M.
c     Note that this vectorizes, and is also fine for cache 
c     based machines.  
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'globals.h'

      integer n1,n2,n3,k
      double precision u(n1,n2,n3),v(n1,n2,n3),r(n1,n2,n3),a(0:3)
!HPF$ template ttt(nz(lt)+2+2**(lt-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with *ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: u, v, r
!HPF$ shadow (0,0,1:1) :: u, v, r
      integer i3, i2, i1
      double precision u1(m), u2(m)

!HPFJ reflect u

!HPF$ independent
      do i3=2,n3-1
!HPF$    on home(r(:,:,i3)), local begin
         do i2=2,n2-1
            do i1=1,n1
               u1(i1) = u(i1,i2-1,i3) + u(i1,i2+1,i3)
     >                + u(i1,i2,i3-1) + u(i1,i2,i3+1)
               u2(i1) = u(i1,i2-1,i3-1) + u(i1,i2+1,i3-1)
     >                + u(i1,i2-1,i3+1) + u(i1,i2+1,i3+1)
            enddo
            do i1=2,n1-1
               r(i1,i2,i3) = v(i1,i2,i3)
     >                     - a(0) * u(i1,i2,i3)
     >                     - a(2) * ( u2(i1) + u1(i1-1) + u1(i1+1) )
     >                     - a(3) * ( u2(i1-1) + u2(i1+1) )
            enddo
         enddo
!HPF$    end on
      enddo

c---------------------------------------------------------------------
c     exchange boundary data
c---------------------------------------------------------------------
      r(n1,:,:) = r(2,:,:)
      r(1,:,:) = r(n1-1,:,:)
      r(:,n2,:) = r(:,2,:)
      r(:,1,:) = r(:,n2-1,:)
      r(:,:,n3) = r(:,:,2)
      r(:,:,1) = r(:,:,n3-1)

      if( debug_vec(0) .ge. 1 )then
         call rep_nrm(r,n1,n2,n3,'   resid',k)
      endif

      if( debug_vec(2) .ge. k )then
         call showall(r,n1,n2,n3,k)
      endif

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine rprj3(r,m1k,m2k,m3k,s,m1j,m2j,m3j,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     rprj3 projects onto the next coarser grid, 
c     using a trilinear Finite Element projection:  s = r' = P r
c     
c     This  implementation costs  20A + 4M per result, where
c     A and M denote the costs of Addition and Multiplication.  
c     Note that this vectorizes, and is also fine for cache 
c     based machines.  
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'globals.h'

      integer m1k, m2k, m3k, m1j, m2j, m3j,k
      double precision r(m1k,m2k,m3k), s(m1j,m2j,m3j)
!HPF$ template ttt(nz(lt)+2+2**(lt-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with *ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: r
!HPF$ align (*,*,i) with *ttt(i*(2**(lt+1-k))-(2**(lt+1-k)-1)) :: s
!HPF$ shadow (0,0,1:1) :: r, s
      integer j3, j2, j1, i3, i2, i1, d1, d2, d3

      double precision x1(m), y1(m), x2,y2


      if(m1k.eq.3)then
        d1 = 2
      else
        d1 = 1
      endif

      if(m2k.eq.3)then
        d2 = 2
      else
        d2 = 1
      endif

      if(m3k.eq.3)then
        d3 = 2
      else
        d3 = 1
      endif

!HPFJ reflect r

!HPF$ independent
      do  j3=2,m3j-1
!HPF$    on home(s(:,:,j3)), local begin
         i3 = 2*j3-d3
         do  j2=2,m2j-1
            i2 = 2*j2-d2

            do j1=2,m1j
              i1 = 2*j1-d1
              x1(i1-1) = r(i1-1,i2-1,i3  ) + r(i1-1,i2+1,i3  )
     >                 + r(i1-1,i2,  i3-1) + r(i1-1,i2,  i3+1)
              y1(i1-1) = r(i1-1,i2-1,i3-1) + r(i1-1,i2-1,i3+1)
     >                 + r(i1-1,i2+1,i3-1) + r(i1-1,i2+1,i3+1)
            enddo

            do  j1=2,m1j-1
              i1 = 2*j1-d1
              y2 = r(i1,  i2-1,i3-1) + r(i1,  i2-1,i3+1)
     >           + r(i1,  i2+1,i3-1) + r(i1,  i2+1,i3+1)
              x2 = r(i1,  i2-1,i3  ) + r(i1,  i2+1,i3  )
     >           + r(i1,  i2,  i3-1) + r(i1,  i2,  i3+1)
              s(j1,j2,j3) =
     >               0.5D0 * r(i1,i2,i3)
     >             + 0.25D0 * ( r(i1-1,i2,i3) + r(i1+1,i2,i3) + x2)
     >             + 0.125D0 * ( x1(i1-1) + x1(i1+1) + y2)
     >             + 0.0625D0 * ( y1(i1-1) + y1(i1+1) )
            enddo

         enddo
!HPF$    end on
      enddo

      s(m1j,:,:) = s(2,:,:)
      s(1,:,:) = s(m1j-1,:,:)
      s(:,m2j,:) = s(:,2,:)
      s(:,1,:) = s(:,m2j-1,:)       
      s(:,:,m3j) = s(:,:,2)
      s(:,:,1) = s(:,:,m3j-1)

      if( debug_vec(0) .ge. 1 )then
         call rep_nrm(s,m1j,m2j,m3j,'   rprj3',k-1)
      endif

      if( debug_vec(4) .ge. k )then
         call showall(s,m1j,m2j,m3j,k-1)
      endif

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine interp( z,mm1,mm2,mm3,u,n1,n2,n3,k )

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     interp adds the trilinear interpolation of the correction
c     from the coarser grid to the current approximation:  u = u + Qu'
c     
c     Observe that this  implementation costs  16A + 4M, where
c     A and M denote the costs of Addition and Multiplication.  
c     Note that this vectorizes, and is also fine for cache 
c     based machines.  Vector machines may get slightly better 
c     performance however, with 8 separate "do i1" loops, rather than 4.
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'globals.h'

      integer mm1, mm2, mm3, n1, n2, n3,k
      double precision z(mm1,mm2,mm3),u(n1,n2,n3)
!HPF$ template ttt(nz(lt)+2+2**(lt-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with *ttt(i*(2**(lt+1-k))-(2**(lt+1-k)-1)) :: z
!HPF$ align (*,*,i) with *ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: u
!HPF$ shadow (0,0,1:1) :: z, u
      integer i3, i2, i1, d1, d2, d3, t1, t2, t3

      double precision z1(m),z2(m),z3(m)

      double precision tmp(mm1,mm2,n3)
!HPF$ align (*,*,i) with ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: tmp
!HPF$ shadow (0,0,1:1) :: tmp

!HPFJ reflect z

      if ( n3 - 2 <= number_of_processors() ) then

!HPF$    independent
         do i3=1, mm3
!HPF$       on home(z(:,:,i3)), local begin
            do i2=1, mm2
               do i1=1, mm1
                  tmp(i1,i2,i3*2-1) = z(i1,i2,i3)
               end do
            end do
!HPF$       end on
         end do

!HPFJ    reflect tmp

!HPF$    independent
         do i3=1, mm3-1
!HPF$       on home(tmp(:,:,i3*2)), local begin
            do i2=1, mm2-1
               do i1=1, mm1-1
                  tmp(i1,i2,i3*2) = tmp(i1,i2,i3*2-1)
               end do
            end do
!HPF$       end on
         end do

!HPF$    independent
         do  i3=1,mm3-1
!HPF$       on home(u(:,:,2*i3-1)), local begin
            do  i2=1,mm2-1

               do i1=1,mm1
                  z1(i1) = z(i1,i2+1,i3) + z(i1,i2,i3)
                  z2(i1) = z(i1,i2,i3+1) + z(i1,i2,i3)
                  z3(i1) = z(i1,i2+1,i3+1) + z(i1,i2,i3+1) + z1(i1)
               enddo

               do  i1=1,mm1-1
                  u(2*i1-1,2*i2-1,2*i3-1)=u(2*i1-1,2*i2-1,2*i3-1)
     >                 +z(i1,i2,i3)
                  u(2*i1,2*i2-1,2*i3-1)=u(2*i1,2*i2-1,2*i3-1)
     >                 +0.5d0*(z(i1+1,i2,i3)+z(i1,i2,i3))
               enddo
               do i1=1,mm1-1
                  u(2*i1-1,2*i2,2*i3-1)=u(2*i1-1,2*i2,2*i3-1)
     >                 +0.5d0 * z1(i1)
                  u(2*i1,2*i2,2*i3-1)=u(2*i1,2*i2,2*i3-1)
     >                 +0.25d0*( z1(i1) + z1(i1+1) )
               enddo
            enddo
!HPF$       end on
         enddo

!HPF$    independent
         do  i3=1,mm3-1
!HPF$       on home(u(:,:,2*i3)), local begin
            do  i2=1,mm2-1

               do i1=1,mm1
                  z1(i1) = tmp(i1,i2+1,2*i3-1) + tmp(i1,i2,2*i3-1)
                  z2(i1) = tmp(i1,i2,2*i3+1) + tmp(i1,i2,2*i3-1)
                  z3(i1) = tmp(i1,i2+1,2*i3+1) + tmp(i1,i2,2*i3+1)
     >                                         + z1(i1)
               enddo

               do i1=1,mm1-1
                  u(2*i1-1,2*i2-1,2*i3)=u(2*i1-1,2*i2-1,2*i3)
     >                 +0.5d0 * z2(i1)
                  u(2*i1,2*i2-1,2*i3)=u(2*i1,2*i2-1,2*i3)
     >                 +0.25d0*( z2(i1) + z2(i1+1) )
               enddo
               do i1=1,mm1-1
                  u(2*i1-1,2*i2,2*i3)=u(2*i1-1,2*i2,2*i3)
     >                 +0.25d0* z3(i1)
                  u(2*i1,2*i2,2*i3)=u(2*i1,2*i2,2*i3)
     >                 +0.125d0*( z3(i1) + z3(i1+1) )
               enddo
            enddo
!HPF$       end on
         enddo

      else if( n1 .ne. 3 .and. n2 .ne. 3 .and. n3 .ne. 3 ) then

!HPF$    independent
         do  i3=1,mm3-1
!HPF$       on home(u(:,:,2*i3-1)), local begin
            do  i2=1,mm2-1

               do i1=1,mm1
                  z1(i1) = z(i1,i2+1,i3) + z(i1,i2,i3)
                  z2(i1) = z(i1,i2,i3+1) + z(i1,i2,i3)
                  z3(i1) = z(i1,i2+1,i3+1) + z(i1,i2,i3+1) + z1(i1)
               enddo

               do  i1=1,mm1-1
                  u(2*i1-1,2*i2-1,2*i3-1)=u(2*i1-1,2*i2-1,2*i3-1)
     >                 +z(i1,i2,i3)
                  u(2*i1,2*i2-1,2*i3-1)=u(2*i1,2*i2-1,2*i3-1)
     >                 +0.5d0*(z(i1+1,i2,i3)+z(i1,i2,i3))
               enddo
               do i1=1,mm1-1
                  u(2*i1-1,2*i2,2*i3-1)=u(2*i1-1,2*i2,2*i3-1)
     >                 +0.5d0 * z1(i1)
                  u(2*i1,2*i2,2*i3-1)=u(2*i1,2*i2,2*i3-1)
     >                 +0.25d0*( z1(i1) + z1(i1+1) )
               enddo
               do i1=1,mm1-1
                  u(2*i1-1,2*i2-1,2*i3)=u(2*i1-1,2*i2-1,2*i3)
     >                 +0.5d0 * z2(i1)
                  u(2*i1,2*i2-1,2*i3)=u(2*i1,2*i2-1,2*i3)
     >                 +0.25d0*( z2(i1) + z2(i1+1) )
               enddo
               do i1=1,mm1-1
                  u(2*i1-1,2*i2,2*i3)=u(2*i1-1,2*i2,2*i3)
     >                 +0.25d0* z3(i1)
                  u(2*i1,2*i2,2*i3)=u(2*i1,2*i2,2*i3)
     >                 +0.125d0*( z3(i1) + z3(i1+1) )
               enddo
            enddo
!HPF$       end on
         enddo

      else

         if(n1.eq.3)then
            d1 = 2
            t1 = 1
         else
            d1 = 1
            t1 = 0
         endif
         
         if(n2.eq.3)then
            d2 = 2
            t2 = 1
         else
            d2 = 1
            t2 = 0
         endif
         
         if(n3.eq.3)then
            d3 = 2
            t3 = 1
         else
            d3 = 1
            t3 = 0
         endif

!HPF$    independent
         do  i3=d3,mm3-1
!HPF$       on home(u(:,:,2*i3-d3)), local begin
            do  i2=d2,mm2-1
               do  i1=d1,mm1-1
                  u(2*i1-d1,2*i2-d2,2*i3-d3)=u(2*i1-d1,2*i2-d2,2*i3-d3)
     >                 +z(i1,i2,i3)
               enddo
               do  i1=1,mm1-1
                  u(2*i1-t1,2*i2-d2,2*i3-d3)=u(2*i1-t1,2*i2-d2,2*i3-d3)
     >                 +0.5D0*(z(i1+1,i2,i3)+z(i1,i2,i3))
               enddo
            enddo
            do  i2=1,mm2-1
               do  i1=d1,mm1-1
                  u(2*i1-d1,2*i2-t2,2*i3-d3)=u(2*i1-d1,2*i2-t2,2*i3-d3)
     >                 +0.5D0*(z(i1,i2+1,i3)+z(i1,i2,i3))
               enddo
               do  i1=1,mm1-1
                  u(2*i1-t1,2*i2-t2,2*i3-d3)=u(2*i1-t1,2*i2-t2,2*i3-d3)
     >                 +0.25D0*(z(i1+1,i2+1,i3)+z(i1+1,i2,i3)
     >                 +z(i1,  i2+1,i3)+z(i1,  i2,i3))
               enddo
            enddo
!HPF$       end on
         enddo

!HPF$    independent
         do  i3=1,mm3-1
!HPF$       on home(u(:,:,2*i3-t3)), local begin
            do  i2=d2,mm2-1
               do  i1=d1,mm1-1
                  u(2*i1-d1,2*i2-d2,2*i3-t3)=u(2*i1-d1,2*i2-d2,2*i3-t3)
     >                 +0.5D0*(z(i1,i2,i3+1)+z(i1,i2,i3))
               enddo
               do  i1=1,mm1-1
                  u(2*i1-t1,2*i2-d2,2*i3-t3)=u(2*i1-t1,2*i2-d2,2*i3-t3)
     >                 +0.25D0*(z(i1+1,i2,i3+1)+z(i1,i2,i3+1)
     >                 +z(i1+1,i2,i3  )+z(i1,i2,i3  ))
               enddo
            enddo
            do  i2=1,mm2-1
               do  i1=d1,mm1-1
                  u(2*i1-d1,2*i2-t2,2*i3-t3)=u(2*i1-d1,2*i2-t2,2*i3-t3)
     >                 +0.25D0*(z(i1,i2+1,i3+1)+z(i1,i2,i3+1)
     >                 +z(i1,i2+1,i3  )+z(i1,i2,i3  ))
               enddo
               do  i1=1,mm1-1
                  u(2*i1-t1,2*i2-t2,2*i3-t3)=u(2*i1-t1,2*i2-t2,2*i3-t3)
     >                 +0.125D0*(z(i1+1,i2+1,i3+1)+z(i1+1,i2,i3+1)
     >                 +z(i1  ,i2+1,i3+1)+z(i1  ,i2,i3+1)
     >                 +z(i1+1,i2+1,i3  )+z(i1+1,i2,i3  )
     >                 +z(i1  ,i2+1,i3  )+z(i1  ,i2,i3  ))
               enddo
            enddo
!HPF$       end on
         enddo

      endif

      if( debug_vec(0) .ge. 1 )then
         call rep_nrm(z,mm1,mm2,mm3,'z: inter',k-1)
         call rep_nrm(u,n1,n2,n3,'u: inter',k)
      endif

      if( debug_vec(5) .ge. k )then
         call showall(z,mm1,mm2,mm3,k-1)
         call showall(u,n1,n2,n3,k)
      endif

      return 
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine norm2u3(r,n1,n2,n3,rnm2,rnmu,nx,ny,nz,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     norm2u3 evaluates approximations to the L2 norm and the
c     uniform (or L-infinity or Chebyshev) norm, under the
c     assumption that the boundaries are periodic or zero.  Add the
c     boundaries in with half weight (quarter weight on the edges
c     and eighth weight at the corners) for inhomogeneous boundaries.
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'npbparams.h'

      integer n1, n2, n3, nx, ny, nz, k
      double precision rnm2, rnmu, r(n1,n2,n3)
!HPF$ template ttt(nz_default+2+2**(lt_default-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with
!HPF$> *ttt(i*(2**(lt_default-k))-(2**(lt_default-k)-1)) :: r
!HPF$ shadow (0,0,1:1) :: r
      double precision s, a, ss
      integer i3, i2, i1, ierr

      integer n

      n = nx*ny*nz

      s=0.0D0
      rnmu = 0.0D0
      do  i3=2,n3-1
         do  i2=2,n2-1
            do  i1=2,n1-1
               s=s+r(i1,i2,i3)**2
               a=abs(r(i1,i2,i3))
               if(a.gt.rnmu)rnmu=a
            enddo
         enddo
      enddo

      rnm2=sqrt( s / float( n ))

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine rep_nrm(u,n1,n2,n3,title,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     report on norm
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'globals.h'

      integer n1, n2, n3, k
      double precision u(n1,n2,n3)
!HPF$ template ttt(nz(lt)+2+2**(lt-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with *ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: u
!HPF$ shadow (0,0,1:1) :: u
      character*8 title

      double precision rnm2, rnmu


      call norm2u3(u,n1,n2,n3,rnm2,rnmu,nx(k),ny(k),nz(k),k)
      write(*,7)k,title,rnm2,rnmu
 7    format(' Level',i2,' in ',a8,': norms =',D21.14,D21.14)

      return
      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine zran3(z,n1,n2,n3,nx,ny,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     zran3  loads +1 at ten randomly chosen points,
c     loads -1 at a different ten random points,
c     and zero elsewhere.
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'npbparams.h'

      integer  is1, is2, is3, ie1, ie2, ie3
      common /grid/ is1,is2,is3,ie1,ie2,ie3

      integer n1, n2, n3, k, nx, ny, ierr, i0, m0, m1
      double precision z(n1,n2,n3)
!HPF$ template ttt(nz_default+2+2**(lt_default-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with
!HPF$> *ttt(i*(2**(lt_default-k))-(2**(lt_default-k)-1)) :: z
!HPF$ shadow (0,0,1:1) :: z

      integer mm, i1, i2, i3, d1, e1, e2, e3
      double precision x, a
      double precision xx, x0, x1, a1, a2, ai, power
      parameter( mm = 10,  a = 5.D0 ** 13, x = 314159265.D0)
      double precision ten( mm, 0:1 ), temp, best
      integer i, j1( mm, 0:1 ), j2( mm, 0:1 ), j3( mm, 0:1 )
      integer jg( 0:3, mm, 0:1 ), jg_temp(4)

      double precision rdummy

      double precision tmp(2:n1)
      logical mask1(n1,n2,n3), mask2(n1,n2,n3)
!HPF$ align (*,*,i) with
!HPF$> ttt(i*(2**(lt_default-k))-(2**(lt_default-k)-1)) :: mask1, mask2
      double precision mx, mn

      interface
         extrinsic(FORTRAN_LOCAL)
     >     SUBROUTINE VRANLC (N, X, A, Y)
         implicit none
         integer n
         double precision x, a, y(*)
         end subroutine

         extrinsic(FORTRAN_LOCAL)
     >     double precision function randlc(x, a)
         implicit none
         double precision x, a
         end function
      end interface


      a1 = power( a, nx )
      a2 = power( a, nx*ny )

      call zero3(z,n1,n2,n3,k)

      i = is1-2+nx*(is2-2+ny*(is3-2))

      ai = power( a, i )
      d1 = ie1 - is1 + 1
      e1 = ie1 - is1 + 2
      e2 = ie2 - is2 + 2
      e3 = ie3 - is3 + 2
      x0 = x
      rdummy = randlc( x0, ai )

      do  i3 = 2, e3
         x1 = x0
         do  i2 = 2, e2
            xx = x1
!            call vranlc( d1, xx, a, z( 2:, i2, i3 ))
            call vranlc( d1, xx, a, tmp)
            z(2:,i2,i3) = tmp
            rdummy = randlc( x1, a1 )
         enddo
         rdummy = randlc( x0, a2 )
      enddo

c---------------------------------------------------------------------
c     each processor looks for twenty candidates
c---------------------------------------------------------------------
!      do  i=1,mm
!         ten( i, 1 ) = 0.0D0
!         j1( i, 1 ) = 0
!         j2( i, 1 ) = 0
!         j3( i, 1 ) = 0
!         ten( i, 0 ) = 1.0D0
!         j1( i, 0 ) = 0
!         j2( i, 0 ) = 0
!         j3( i, 0 ) = 0
!      enddo
!
!!HPF$ independent
!      do  i3=2,n3-1
!!HPF$    on home(z(:,:,i3)), local begin
!         do  i2=2,n2-1
!            do  i1=2,n1-1
!               if( z(i1,i2,i3) .gt. ten( 1, 1 ) )then
!                  ten(1,1) = z(i1,i2,i3) 
!                  j1(1,1) = i1
!                  j2(1,1) = i2
!                  j3(1,1) = i3
!                  call bubble( ten, j1, j2, j3, mm, 1 )
!               endif
!               if( z(i1,i2,i3) .lt. ten( 1, 0 ) )then
!                  ten(1,0) = z(i1,i2,i3) 
!                  j1(1,0) = i1
!                  j2(1,0) = i2
!                  j3(1,0) = i3
!                  call bubble( ten, j1, j2, j3, mm, 0 )
!               endif
!            enddo
!         enddo
!!HPF$    end on
!      enddo

      mask1 = .true.
      mask2 = .true.

      do i=1, mm

         mx = 0.0D0
         mn = 1.0D0

!HPF$    independent, reduction(MAX:mx), reduction(MIN:mn)
         do i3=2, n3-1
            do i2=2, n2-1
               do i1=2, n1-1
                  if (mask1(i1,i2,i3) .and. mask2(i1,i2,i3)) then
                     mx = max(z(i1,i2,i3), mx)
                     mn = min(z(i1,i2,i3), mn)
                  end if
               end do
            end do
         end do

         do i3=2, n3-1
            do i2=2, n2-1
               do i1=2, n1-1
                  if (z(i1,i2,i3) == mx) then
                     mask1(i1,i2,i3) = .false.
                  else if (z(i1,i2,i3) == mn) then
                     mask2(i1,i2,i3) = .false.
                  end if
               end do
            end do
         end do

      end do

c---------------------------------------------------------------------
c     Now which of these are globally best?
c---------------------------------------------------------------------
!      i1 = mm
!      i0 = mm
!
!      do  i=mm,1,-1
!
!         best = z( j1(i1,1), j2(i1,1), j3(i1,1) )
!         if(best.eq.z(j1(i1,1),j2(i1,1),j3(i1,1)))then
!            jg( 0, i, 1) = 0
!            jg( 1, i, 1) = is1 - 2 + j1( i1, 1 ) 
!            jg( 2, i, 1) = is2 - 2 + j2( i1, 1 ) 
!            jg( 3, i, 1) = is3 - 2 + j3( i1, 1 ) 
!            i1 = i1-1
!         else
!            jg( 0, i, 1) = 0
!            jg( 1, i, 1) = 0
!            jg( 2, i, 1) = 0
!            jg( 3, i, 1) = 0
!         endif
!         ten( i, 1 ) = best
!
!         best = z( j1(i0,0), j2(i0,0), j3(i0,0) )
!         if(best.eq.z(j1(i0,0),j2(i0,0),j3(i0,0)))then
!            jg( 0, i, 0) = 0
!            jg( 1, i, 0) = is1 - 2 + j1( i0, 0 ) 
!            jg( 2, i, 0) = is2 - 2 + j2( i0, 0 ) 
!            jg( 3, i, 0) = is3 - 2 + j3( i0, 0 ) 
!            i0 = i0-1
!         else
!            jg( 0, i, 0) = 0
!            jg( 1, i, 0) = 0
!            jg( 2, i, 0) = 0
!            jg( 3, i, 0) = 0
!         endif
!         ten( i, 0 ) = best
!
!      enddo

!      m1 = i1+1
!      m0 = i0+1

C      write(*,*)' '
C      write(*,*)' negative charges at'
C      write(*,9)(jg(1,i,0),jg(2,i,0),jg(3,i,0),i=1,mm)
C      write(*,*)' positive charges at'
C      write(*,9)(jg(1,i,1),jg(2,i,1),jg(3,i,1),i=1,mm)
C      write(*,*)' small random numbers were'
C      write(*,8)(ten( i,0),i=mm,1,-1)
C      write(*,*)' and they were found on processor number'
C      write(*,7)(jg(0,i,0),i=mm,1,-1)
C      write(*,*)' large random numbers were'
C      write(*,8)(ten( i,1),i=mm,1,-1)
C      write(*,*)' and they were found on processor number'
C      write(*,7)(jg(0,i,1),i=mm,1,-1)
C 9    format(5(' (',i3,2(',',i3),')'))
C 8    format(5D15.8)
C 7    format(10i4)
!      do  i3=1,n3
!         do  i2=1,n2
!            do  i1=1,n1
!               z(i1,i2,i3) = 0.0D0
!            enddo
!         enddo
!      enddo
!      do  i=mm,m0,-1
!         z( j1(i,0), j2(i,0), j3(i,0) ) = -1.0D0
!      enddo
!      do  i=mm,m1,-1
!         z( j1(i,1), j2(i,1), j3(i,1) ) = +1.0D0
!      enddo
      do  i3=2,n3-1
         do  i2=2,n2-1
            do  i1=2,n1-1
               if (.not.mask1(i1,i2,i3)) then
                  z(i1,i2,i3) = +1.0D0
               else if (.not.mask2(i1,i2,i3)) then
                  z(i1,i2,i3) = -1.0D0
               else
                  z(i1,i2,i3) = 0.0D0
               end if
            enddo
         enddo
      enddo

      z(n1,:,:) = z(2,:,:)
      z(1,:,:) = z(n1-1,:,:)
      z(:,n2,:) = z(:,2,:)
      z(:,1,:) = z(:,n2-1,:)       
      z(:,:,n3) = z(:,:,2)
      z(:,:,1) = z(:,:,n3-1)

c---------------------------------------------------------------------
c          call showall(z,n1,n2,n3)
c---------------------------------------------------------------------

      return 
      end


c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine showall(z,n1,n2,n3,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      use mapping

      implicit none

      include 'npbparams.h'

      integer n1,n2,n3,i1,i2,i3,i,ierr,k
      double precision z(n1,n2,n3)
!HPF$ template ttt(nz_default+2+2**(lt_default-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with
!HPF$> *ttt(i*(2**(lt_default-k))-(2**(lt_default-k)-1)) :: z
!HPF$ shadow (0,0,1:1) :: z
      integer m1, m2, m3

      m1 = min(n1,18)
      m2 = min(n2,14)
      m3 = min(n3,18)

      write(*,*)'  '
      do  i3=1,m3
         do  i1=1,m1
            write(*,6)(z(i1,i2,i3),i2=1,m2)
         enddo
         write(*,*)' - - - - - - - '
      enddo
      write(*,*)'  '
 6    format(15f6.3)

      return 
      end


c---------------------------------------------------------------------
c---------------------------------------------------------------------

      double precision function power( a, n )

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c     power  raises an integer, disguised as a double
c     precision real, to an integer power
c---------------------------------------------------------------------
      implicit none

      double precision a, aj
      integer n, nj
      external randlc
      double precision randlc, rdummy

      power = 1.0D0
      nj = n
      aj = a
 100  continue

      if( nj .eq. 0 ) goto 200
      if( mod(nj,2) .eq. 1 ) rdummy =  randlc( power, aj )
      rdummy = randlc( aj, aj )
      nj = nj/2
      go to 100

 200  continue
      return
      end

!c---------------------------------------------------------------------
!c---------------------------------------------------------------------
!
!      pure subroutine bubble( ten, j1, j2, j3, m, ind )
!
!c---------------------------------------------------------------------
!c---------------------------------------------------------------------
!
!c---------------------------------------------------------------------
!c     bubble        does a bubble sort in direction dir
!c---------------------------------------------------------------------
!      implicit none
!
!      integer m, ind, j1( m, 0:1 ), j2( m, 0:1 ), j3( m, 0:1 )
!      double precision ten( m, 0:1 )
!      double precision temp
!      integer i, j_temp
!
!      if( ind .eq. 1 )then
!
!         do  i=1,m-1
!            if( ten(i,ind) .gt. ten(i+1,ind) )then
!
!               temp = ten( i+1, ind )
!               ten( i+1, ind ) = ten( i, ind )
!               ten( i, ind ) = temp
!
!               j_temp           = j1( i+1, ind )
!               j1( i+1, ind ) = j1( i,   ind )
!               j1( i,   ind ) = j_temp
!
!               j_temp           = j2( i+1, ind )
!               j2( i+1, ind ) = j2( i,   ind )
!               j2( i,   ind ) = j_temp
!
!               j_temp           = j3( i+1, ind )
!               j3( i+1, ind ) = j3( i,   ind )
!               j3( i,   ind ) = j_temp
!
!            else 
!               return
!            endif
!         enddo
!
!      else
!
!         do  i=1,m-1
!            if( ten(i,ind) .lt. ten(i+1,ind) )then
!
!               temp = ten( i+1, ind )
!               ten( i+1, ind ) = ten( i, ind )
!               ten( i, ind ) = temp
!
!               j_temp           = j1( i+1, ind )
!               j1( i+1, ind ) = j1( i,   ind )
!               j1( i,   ind ) = j_temp
!
!               j_temp           = j2( i+1, ind )
!               j2( i+1, ind ) = j2( i,   ind )
!               j2( i,   ind ) = j_temp
!
!               j_temp           = j3( i+1, ind )
!               j3( i+1, ind ) = j3( i,   ind )
!               j3( i,   ind ) = j_temp
!
!            else 
!               return
!            endif
!         enddo
!
!      endif
!
!      return
!      end

c---------------------------------------------------------------------
c---------------------------------------------------------------------

      subroutine zero3(z,n1,n2,n3,k)

c---------------------------------------------------------------------
c---------------------------------------------------------------------
      use mapping

      implicit none

      include 'globals.h'

      integer n1, n2, n3, k
      double precision z(n1,n2,n3)
!HPF$ template ttt(nz(lt)+2+2**(lt-1))
!HPF$ distribute (gen_block(map)) :: ttt
!HPF$ align (*,*,i) with *ttt(i*(2**(lt-k))-(2**(lt-k)-1)) :: z
!HPF$ shadow (0,0,1:1) :: z
      integer i1, i2, i3

      do  i3=1,n3
         do  i2=1,n2
            do  i1=1,n1
               z(i1,i2,i3)=0.0D0
            enddo
         enddo
      enddo

      return
      end


c----- end of program ------------------------------------------------
