
c---------------------------------------------------------------------
c---------------------------------------------------------------------

       subroutine exact_rhs

c---------------------------------------------------------------------
c---------------------------------------------------------------------

c---------------------------------------------------------------------
c compute the right hand side based on exact solution
c---------------------------------------------------------------------

       include 'header.h'

       double precision dtemp(5), xi, eta, zeta, dtpp
!HPF$  sequence dtemp
       integer          c, m, i, j, k, ip1, im1, jp1, 
     >                  jm1, km1, kp1

      interface
         pure extrinsic(FORTRAN_LOCAL)
     >     subroutine exact_solution(xi,eta,zeta,dtemp)
         double precision  xi, eta, zeta, dtemp(5)
!HPF$    sequence dtemp
         end subroutine
      end interface

       c = 1

c---------------------------------------------------------------------
c      initialize                                  
c---------------------------------------------------------------------
       do   m = 1, 5
          do   k= 0, grid_points(3)-1
             do   j = 0, grid_points(2)-1
                do   i = 0, grid_points(1)-1
                   forcing(i,j,k,m) = 0.0d0
                end do
             end do
          end do
       end do

c---------------------------------------------------------------------
c      xi-direction flux differences                      
c---------------------------------------------------------------------
!HPF$  independent
       do   k = 1, grid_points(3)-2
          zeta = dble(k) * dnzm1
          do   j = 1, grid_points(2)-2
             eta = dble(j) * dnym1

             do  i=0, grid_points(1)-1
                xi = dble(i) * dnxm1

                call exact_solution(xi, eta, zeta, dtemp)
                do  m = 1, 5
                   ue2(i,m) = dtemp(m)
                end do

                dtpp = 1.0d0 / dtemp(1)

                do  m = 2, 5
                   buf2(i,m) = dtpp * dtemp(m)
                end do

                cuf2(i)   = buf2(i,2) * buf2(i,2)
                buf2(i,1) = cuf2(i) + buf2(i,3) * buf2(i,3) + 
     >                     buf2(i,4) * buf2(i,4) 
                q2(i) = 0.5d0*(buf2(i,2)*ue2(i,2) + buf2(i,3)*ue2(i,3) +
     >                        buf2(i,4)*ue2(i,4))

             end do
 
             do  i = 1, grid_points(1)-2
                im1 = i-1
                ip1 = i+1

                forcing(i,j,k,1) = forcing(i,j,k,1) -
     >                 tx2*( ue2(ip1,2)-ue2(im1,2) )+
     >                 dx1tx1*(ue2(ip1,1)-2.0d0*ue2(i,1)+ue2(im1,1))

                forcing(i,j,k,2) = forcing(i,j,k,2) - tx2 * (
     >                (ue2(ip1,2)*buf2(ip1,2)+c2*(ue2(ip1,5)-q2(ip1)))-
     >                (ue2(im1,2)*buf2(im1,2)+c2*(ue2(im1,5)-q2(im1))))+
     >                 xxcon1*(buf2(ip1,2)-2.0d0*buf2(i,2)+buf2(im1,2))+
     >                 dx2tx1*( ue2(ip1,2)-2.0d0* ue2(i,2)+ue2(im1,2))

                forcing(i,j,k,3) = forcing(i,j,k,3) - tx2 * (
     >                 ue2(ip1,3)*buf2(ip1,2)-ue2(im1,3)*buf2(im1,2))+
     >                 xxcon2*(buf2(ip1,3)-2.0d0*buf2(i,3)+buf2(im1,3))+
     >                 dx3tx1*( ue2(ip1,3)-2.0d0*ue2(i,3) +ue2(im1,3))
                  
                forcing(i,j,k,4) = forcing(i,j,k,4) - tx2*(
     >                 ue2(ip1,4)*buf2(ip1,2)-ue2(im1,4)*buf2(im1,2))+
     >                 xxcon2*(buf2(ip1,4)-2.0d0*buf2(i,4)+buf2(im1,4))+
     >                 dx4tx1*( ue2(ip1,4)-2.0d0* ue2(i,4)+ ue2(im1,4))

                forcing(i,j,k,5) = forcing(i,j,k,5) - tx2*(
     >                 buf2(ip1,2)*(c1*ue2(ip1,5)-c2*q2(ip1))-
     >                 buf2(im1,2)*(c1*ue2(im1,5)-c2*q2(im1)))+
     >                 0.5d0*xxcon3*(buf2(ip1,1)-2.0d0*buf2(i,1)+
     >                               buf2(im1,1))+
     >                 xxcon4*(cuf2(ip1)-2.0d0*cuf2(i)+cuf2(im1))+
     >                 xxcon5*(buf2(ip1,5)-2.0d0*buf2(i,5)+buf2(im1,5))+
     >                 dx5tx1*( ue2(ip1,5)-2.0d0* ue2(i,5)+ ue2(im1,5))
             end do

c---------------------------------------------------------------------
c            Fourth-order dissipation                         
c---------------------------------------------------------------------
             do   m = 1, 5
                i = 1
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                  (5.0d0*ue2(i,m) - 4.0d0*ue2(i+1,m) +ue2(i+2,m))
                i = 2
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                   (-4.0d0*ue2(i-1,m) + 6.0d0*ue2(i,m) -
     >                     4.0d0*ue2(i+1,m) +       ue2(i+2,m))
             end do

             do   m = 1, 5
                do  i = 3, grid_points(1)-4
                   forcing(i,j,k,m) = forcing(i,j,k,m) - dssp*
     >                   (ue2(i-2,m) - 4.0d0*ue2(i-1,m) +
     >                  6.0d0*ue2(i,m) - 4.0d0*ue2(i+1,m) + ue2(i+2,m))
                end do
             end do

             do   m = 1, 5
                i = grid_points(1)-3
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                   (ue2(i-2,m) - 4.0d0*ue2(i-1,m) +
     >                    6.0d0*ue2(i,m) - 4.0d0*ue2(i+1,m))
                i = grid_points(1)-2
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                 (ue2(i-2,m) - 4.0d0*ue2(i-1,m) + 5.0d0*ue2(i,m))
             end do

          end do
       end do

c---------------------------------------------------------------------
c  eta-direction flux differences             
c---------------------------------------------------------------------
!HPF$  independent
       do   k = 1, grid_points(3)-2          
          zeta = dble(k) * dnzm1
          do   i=1, grid_points(1)-2
             xi = dble(i) * dnxm1

             do  j=0, grid_points(2)-1
                eta = dble(j) * dnym1

                call exact_solution(xi, eta, zeta, dtemp)
                do   m = 1, 5 
                   ue2(j,m) = dtemp(m)
                end do
                dtpp = 1.0d0/dtemp(1)

                do  m = 2, 5
                   buf2(j,m) = dtpp * dtemp(m)
                end do

                cuf2(j)   = buf2(j,3) * buf2(j,3)
                buf2(j,1) = cuf2(j) + buf2(j,2) * buf2(j,2) + 
     >                     buf2(j,4) * buf2(j,4)
                q2(j) = 0.5d0*(buf2(j,2)*ue2(j,2) + buf2(j,3)*ue2(j,3) +
     >                        buf2(j,4)*ue2(j,4))
             end do

             do  j = 1, grid_points(2)-2
                jm1 = j-1
                jp1 = j+1
                  
                forcing(i,j,k,1) = forcing(i,j,k,1) -
     >                ty2*( ue2(jp1,3)-ue2(jm1,3) )+
     >                dy1ty1*(ue2(jp1,1)-2.0d0*ue2(j,1)+ue2(jm1,1))

                forcing(i,j,k,2) = forcing(i,j,k,2) - ty2*(
     >                ue2(jp1,2)*buf2(jp1,3)-ue2(jm1,2)*buf2(jm1,3))+
     >                yycon2*(buf2(jp1,2)-2.0d0*buf2(j,2)+buf2(jm1,2))+
     >                dy2ty1*( ue2(jp1,2)-2.0* ue2(j,2)+ ue2(jm1,2))

                forcing(i,j,k,3) = forcing(i,j,k,3) - ty2*(
     >                (ue2(jp1,3)*buf2(jp1,3)+c2*(ue2(jp1,5)-q2(jp1)))-
     >                (ue2(jm1,3)*buf2(jm1,3)+c2*(ue2(jm1,5)-q2(jm1))))+
     >                yycon1*(buf2(jp1,3)-2.0d0*buf2(j,3)+buf2(jm1,3))+
     >                dy3ty1*( ue2(jp1,3)-2.0d0*ue2(j,3) +ue2(jm1,3))

                forcing(i,j,k,4) = forcing(i,j,k,4) - ty2*(
     >                ue2(jp1,4)*buf2(jp1,3)-ue2(jm1,4)*buf2(jm1,3))+
     >                yycon2*(buf2(jp1,4)-2.0d0*buf2(j,4)+buf2(jm1,4))+
     >                dy4ty1*( ue2(jp1,4)-2.0d0*ue2(j,4)+ ue2(jm1,4))

                forcing(i,j,k,5) = forcing(i,j,k,5) - ty2*(
     >                buf2(jp1,3)*(c1*ue2(jp1,5)-c2*q2(jp1))-
     >                buf2(jm1,3)*(c1*ue2(jm1,5)-c2*q2(jm1)))+
     >                0.5d0*yycon3*(buf2(jp1,1)-2.0d0*buf2(j,1)+
     >                              buf2(jm1,1))+
     >                yycon4*(cuf2(jp1)-2.0d0*cuf2(j)+cuf2(jm1))+
     >                yycon5*(buf2(jp1,5)-2.0d0*buf2(j,5)+buf2(jm1,5))+
     >                dy5ty1*(ue2(jp1,5)-2.0d0*ue2(j,5)+ue2(jm1,5))
             end do

c---------------------------------------------------------------------
c            Fourth-order dissipation                      
c---------------------------------------------------------------------
             do   m = 1, 5
                j = 1
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                  (5.0d0*ue2(j,m) - 4.0d0*ue2(j+1,m) +ue2(j+2,m))
                j = 2
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                   (-4.0d0*ue2(j-1,m) + 6.0d0*ue2(j,m) -
     >                     4.0d0*ue2(j+1,m) +       ue2(j+2,m))
             end do

             do   m = 1, 5
                do  j = 3, grid_points(2)-4
                   forcing(i,j,k,m) = forcing(i,j,k,m) - dssp*
     >                   (ue2(j-2,m) - 4.0d0*ue2(j-1,m) +
     >                  6.0d0*ue2(j,m) - 4.0d0*ue2(j+1,m) + ue2(j+2,m))
                end do
             end do

             do   m = 1, 5
                j = grid_points(2)-3
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                   (ue2(j-2,m) - 4.0d0*ue2(j-1,m) +
     >                    6.0d0*ue2(j,m) - 4.0d0*ue2(j+1,m))
                j = grid_points(2)-2
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                 (ue2(j-2,m) - 4.0d0*ue2(j-1,m) + 5.0d0*ue2(j,m))

             end do

          end do
       end do

c---------------------------------------------------------------------
c      zeta-direction flux differences                      
c---------------------------------------------------------------------
       do  j=1, grid_points(2)-2
          eta = dble(j) * dnym1
          do   i = 1, grid_points(1)-2
             xi = dble(i) * dnxm1

             do k=0, grid_points(3)-1
                zeta = dble(k) * dnzm1

                call exact_solution(xi, eta, zeta, dtemp)
                do   m = 1, 5
                   ue(k,m) = dtemp(m)
                end do

                dtpp = 1.0d0/dtemp(1)

                do   m = 2, 5
                   buf(k,m) = dtpp * dtemp(m)
                end do

                cuf(k)   = buf(k,4) * buf(k,4)
                buf(k,1) = cuf(k) + buf(k,2) * buf(k,2) + 
     >                     buf(k,3) * buf(k,3)
                q(k) = 0.5d0*(buf(k,2)*ue(k,2) + buf(k,3)*ue(k,3) +
     >                        buf(k,4)*ue(k,4))
             end do

             do    k=1, grid_points(3)-2
                km1 = k-1
                kp1 = k+1
                  
                forcing(i,j,k,1) = forcing(i,j,k,1) -
     >                 tz2*( ue(kp1,4)-ue(km1,4) )+
     >                 dz1tz1*(ue(kp1,1)-2.0d0*ue(k,1)+ue(km1,1))

                forcing(i,j,k,2) = forcing(i,j,k,2) - tz2 * (
     >                 ue(kp1,2)*buf(kp1,4)-ue(km1,2)*buf(km1,4))+
     >                 zzcon2*(buf(kp1,2)-2.0d0*buf(k,2)+buf(km1,2))+
     >                 dz2tz1*( ue(kp1,2)-2.0d0* ue(k,2)+ ue(km1,2))

                forcing(i,j,k,3) = forcing(i,j,k,3) - tz2 * (
     >                 ue(kp1,3)*buf(kp1,4)-ue(km1,3)*buf(km1,4))+
     >                 zzcon2*(buf(kp1,3)-2.0d0*buf(k,3)+buf(km1,3))+
     >                 dz3tz1*(ue(kp1,3)-2.0d0*ue(k,3)+ue(km1,3))

                forcing(i,j,k,4) = forcing(i,j,k,4) - tz2 * (
     >                (ue(kp1,4)*buf(kp1,4)+c2*(ue(kp1,5)-q(kp1)))-
     >                (ue(km1,4)*buf(km1,4)+c2*(ue(km1,5)-q(km1))))+
     >                zzcon1*(buf(kp1,4)-2.0d0*buf(k,4)+buf(km1,4))+
     >                dz4tz1*( ue(kp1,4)-2.0d0*ue(k,4) +ue(km1,4))

                forcing(i,j,k,5) = forcing(i,j,k,5) - tz2 * (
     >                 buf(kp1,4)*(c1*ue(kp1,5)-c2*q(kp1))-
     >                 buf(km1,4)*(c1*ue(km1,5)-c2*q(km1)))+
     >                 0.5d0*zzcon3*(buf(kp1,1)-2.0d0*buf(k,1)
     >                              +buf(km1,1))+
     >                 zzcon4*(cuf(kp1)-2.0d0*cuf(k)+cuf(km1))+
     >                 zzcon5*(buf(kp1,5)-2.0d0*buf(k,5)+buf(km1,5))+
     >                 dz5tz1*( ue(kp1,5)-2.0d0*ue(k,5)+ ue(km1,5))
             end do

c---------------------------------------------------------------------
c            Fourth-order dissipation                        
c---------------------------------------------------------------------

!HPFJ        reflect ue

             do   m = 1, 5
                k = 1
!HPF$           on home(forcing(i,j,k,m)), local begin
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                    (5.0d0*ue(k,m) - 4.0d0*ue(k+1,m) +ue(k+2,m))
!HPF$           end on
                k = 2
!HPF$           on home(forcing(i,j,k,m)), local begin
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                   (-4.0d0*ue(k-1,m) + 6.0d0*ue(k,m) -
     >                     4.0d0*ue(k+1,m) +       ue(k+2,m))
!HPF$           end on
             end do

             do   m = 1, 5
!HPF$           independent
                do  k = 3, grid_points(3)-4
!HPF$              on home(forcing(i,j,k,m)) begin
                   forcing(i,j,k,m) = forcing(i,j,k,m) - dssp*
     >                   (ue(k-2,m) - 4.0d0*ue(k-1,m) +
     >                    6.0d0*ue(k,m) - 4.0d0*ue(k+1,m) + ue(k+2,m))
!HPF$              end on
                end do
             end do

             do    m = 1, 5
                k = grid_points(3)-3
!HPF$           on home(forcing(i,j,k,m)), local begin
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                   (ue(k-2,m) - 4.0d0*ue(k-1,m) +
     >                    6.0d0*ue(k,m) - 4.0d0*ue(k+1,m))
!HPF$           end on
                k = grid_points(3)-2
!HPF$           on home(forcing(i,j,k,m)), local begin
                forcing(i,j,k,m) = forcing(i,j,k,m) - dssp *
     >                   (ue(k-2,m) - 4.0d0*ue(k-1,m) + 5.0d0*ue(k,m))
!HPF$           end on
             end do

          end do
       end do

c---------------------------------------------------------------------
c now change the sign of the forcing function, 
c---------------------------------------------------------------------
       do   m = 1, 5
          do   k = 1, grid_points(3)-2
             do   j = 1, grid_points(2)-2
                do   i = 1, grid_points(1)-2
                   forcing(i,j,k,m) = -1.d0 * forcing(i,j,k,m)
                end do
             end do
          end do
       end do

       return
       end





